<?php
/**
*  2020 - Contabilium.com
*
*  @author    Federico Ferreri <federico@ferreri.com.ar>
*  @copyright Contabilium S.A.
*/

if (!defined('_PS_VERSION_')) {
    exit;
}

/**
 * Incluyo el conector a Contabilium
 */
require_once(dirname(__FILE__) . '/lib/Connector.php');

class Contabilium extends Module
{
    protected $config_form = false;
    protected $contabilium = null;
    protected $is_presta_17 = false;


    /**
     * Constructor del modulo.
     */
    public function __construct()
    {
        $this->name = 'contabilium';
        $this->tab = 'billing_invoicing';
        $this->version = '2.1.2';
        $this->author = 'Contabilium.com';
        $this->need_instance = 1;

        $this->bootstrap = true;

        parent::__construct();

        $this->displayName = $this->l('Contabilium.com Integration Module');
        $this->description = $this->l('Allows the store owner to integrate this PrestaShop site with the cloud accounting software Contabilium.com');
        $this->confirmUninstall = $this->l('Are you sure? After uninstalling this module Contabilium.com integration will no longer work.');

        $this->ps_versions_compliancy = array('min' => '1.6', 'max' => _PS_VERSION_);
        $this->is_presta_17 = version_compare(_PS_VERSION_, '1.7.0.0') >= 0;

        $this->contabilium = new Connector(Configuration::get('CONTABILIUM_USERNAME'), Configuration::get('CONTABILIUM_API_KEY'), Configuration::get('CONTABILIUM_COUNTRY_CODE'));
    }

    public function install($update_params = true)
    {
        if ($update_params) {
            Configuration::updateValue('CONTABILIUM_LIVE_MODE', 0);
            Configuration::updateValue('CONTABILIUM_COUNTRY_CODE', '');
            Configuration::updateValue('CONTABILIUM_USERNAME', null);
            Configuration::updateValue('CONTABILIUM_API_KEY', null);
            Configuration::updateValue('CONTABILIUM_INTEGRATION_ID', null);
            Configuration::updateValue('CONTABILIUM_SYNC_PRICE', 1);
            Configuration::updateValue('CONTABILIUM_USE_PRICE_WITH_TAX', Configuration::get('PS_TAX'));
            Configuration::updateValue('CONTABILIUM_SYNC_STOCK', 1);
            Configuration::updateValue('CONTABILIUM_ACCEPTED_STATUS', serialize(array(Configuration::get('PS_OS_PREPARATION'))));
            Configuration::updateValue('CONTABILIUM_CANCELED_STATUS', serialize(array(Configuration::get('PS_OS_CANCELED'))));
            Configuration::updateValue('CONTABILIUM_CALLBACK_TOKEN', ($this->is_presta_17) ? Tools::hash(microtime() . print_r($_SERVER, true)) : Tools::encrypt(microtime() . print_r($_SERVER, true)));
            Configuration::updateValue('CONTABILIUM_TOKEN_CACHE', null);
        }

        include(dirname(__FILE__).'/sql/install.php');

        return parent::install() &&
            $this->registerHook('header') &&
            $this->registerHook('backOfficeHeader') &&
            $this->registerHook('actionOrderStatusPostUpdate') &&
            $this->registerHook('actionOrderEdited') &&
            $this->registerHook('displayAdminProductsExtra') &&
            $this->registerHook('displayAdminOrder');
    }

    public function uninstall($delete_params = true)
    {
        if ($delete_params) {
            Configuration::deleteByName('CONTABILIUM_LIVE_MODE');
            Configuration::deleteByName('CONTABILIUM_USERNAME');
            Configuration::deleteByName('CONTABILIUM_API_KEY');
            Configuration::deleteByName('CONTABILIUM_INTEGRATION_ID');
            Configuration::deleteByName('CONTABILIUM_SYNC_PRICE');
            Configuration::deleteByName('CONTABILIUM_USE_PRICE_WITH_TAX');
            Configuration::deleteByName('CONTABILIUM_SYNC_STOCK');
            Configuration::deleteByName('CONTABILIUM_ACCEPTED_STATUS');
            Configuration::deleteByName('CONTABILIUM_CANCELED_STATUS');
            Configuration::deleteByName('CONTABILIUM_CALLBACK_TOKEN');
            Configuration::deleteByName('CONTABILIUM_TOKEN_CACHE');
        }

        return parent::uninstall();
    }

    public function reset()
    {
        if (!$this->uninstall(false)) {
            return false;
        }
        if (!$this->install(false)) {
            return false;
        }

        return true;
    }

    public function getContent()
    {
        if (((bool)Tools::isSubmit('submitContabiliumModule')) == true) {
            if (Tools::getValue('submitContabiliumSyncNow') == 'X') {
                if (!$this->contabilium->triggerManualSync()->hasError()) {
                    if (!$this->contabilium->wasQueued) {
                        $this->adminDisplayInformation($this->l('Manual synchronization succeded!') . ' - ' . sprintf($this->l('%d product(s) updated.'), $this->contabilium->getUpdatedProducts(true)));
                    } else {
                        $this->adminDisplayInformation($this->l('Manual synchronization started! A process will run in the background an all matching products will be uptated.'));
                    }
                } else {
                    $this->adminDisplayWarning('<b>'. $this->l('Something went wrong') . ':</b> ' . $this->contabilium->getLastError());
                }
            } elseif (substr(Tools::getValue('submitContabiliumCheckItNow'), 0, 2) == 'CS') {
                $code = trim(substr(Tools::getValue('submitContabiliumCheckItNow'), 3));
                if (!$this->contabilium->triggerSync($code)->hasError()) {
                    $message = "<ul>" . array_reduce($this->contabilium->getMessages(), function($carry, $item) { return $carry . "<li>" . $item . "</li>"; }) . "</ul>";
                    $this->adminDisplayInformation('<b>'. $this->l('Test synchronization succeded!') . '</b>' . $message);
                } else {
                    $this->adminDisplayWarning('<b>'. $this->l('Something went wrong') . ':</b> ' . $this->contabilium->getLastError());
                }
            } elseif (Tools::getValue('submitContabiliumRegenerateToken') == 'R') {
                Configuration::updateValue('CONTABILIUM_CALLBACK_TOKEN', ($this->is_presta_17) ? Tools::hash(microtime() . print_r($_SERVER, true)) : Tools::encrypt(microtime() . print_r($_SERVER, true)));
                $this->adminDisplayInformation($this->l('Callback URL regeneration DONE! You must notify CONTABILIUM about this change. Otherwise, integration will not work correctly.'));
            } elseif (Tools::getValue('submitContabiliumModule') == 1) {
                $this->postProcessSave();
                $this->contabilium = new Connector(Configuration::get('CONTABILIUM_USERNAME'), Configuration::get('CONTABILIUM_API_KEY'), Configuration::get('CONTABILIUM_COUNTRY_CODE'));
                $this->contabilium->connect(true);
            }
        }

        $this->context->smarty->assign('module_dir', $this->_path);
        $output = $this->context->smarty->fetch($this->local_path.'views/templates/admin/configure.tpl');
        return $output.$this->renderForm();
    }

    protected function renderForm()
    {
        $helper = new HelperForm();

        $helper->show_toolbar = false;
        $helper->table = $this->table;
        $helper->module = $this;
        $helper->default_form_language = $this->context->language->id;
        $helper->allow_employee_form_lang = Configuration::get('PS_BO_ALLOW_EMPLOYEE_FORM_LANG', 0);

        $helper->identifier = $this->identifier;
        $helper->submit_action = 'submitContabiliumModule';
        $helper->currentIndex = $this->context->link->getAdminLink('AdminModules', false).'&configure='.$this->name.'&tab_module='.$this->tab.'&module_name='.$this->name;
        $helper->token = Tools::getAdminTokenLite('AdminModules');

        $helper->tpl_vars = array(
            'fields_value' => $this->getConfigFormValues(),
            'languages' => $this->context->controller->getLanguages(),
            'id_language' => $this->context->language->id,
        );

        if (Shop::getContext() == Shop::CONTEXT_SHOP) {
            return $helper->generateForm(array($this->getConfigForm()));
        } else {
            return '<div class="row alert alert-warning"><div class="col-xs-12 text-left"><b>' . $this->l('Only valid for Shops Context. Please select a shop from the menu at the top.') . '</div></div>';
        }
    }

    protected function getConfigForm()
    {
        $orderStates = OrderState::getOrderStates($this->context->employee->id_lang);
        $os = array();
        foreach ($orderStates as $value) {
            $os[] = array('id_option' => $value['id_order_state'], 'name' => $value['name']);
        }

        $thisShop = new Shop(Shop::getContextShopID());
        $base_url = Tools::substr($thisShop->getBaseUrl(true, true), 0, -1) . Tools::str_replace_once(_PS_ROOT_DIR_, '', $this->getLocalPath());

        $this->contabilium->connect();

        $form = array(
            'form' => array(
                'legend' => array(
                    'title' => $this->l('Contabilium.com integration settings'),
                    'icon' => 'icon-cogs',
                ),
                'input' => array(
                    array(
                        'type' => 'switch',
                        'label' => $this->l('Enabled'),
                        'name' => 'CONTABILIUM_LIVE_MODE',
                        'is_bool' => true,
                        'desc' => $this->l('Enable integration between Contabilim.com and this PrestaShop site.'),
                        'values' => array(
                            array(
                                'id' => 'active_on',
                                'value' => 1,
                                'label' => $this->l('Enabled')
                            ),
                            array(
                                'id' => 'active_off',
                                'value' => 0,
                                'label' => $this->l('Disabled')
                            )
                        ),
                    ),

                    array(
                        'type' => 'select',
                        'multiple' => false,
                        'name' => 'CONTABILIUM_COUNTRY_CODE',
                        'label' => $this->l('Country'),
                        'required' => true,
                        'options' => array(
                            'query' => array(
                                array(
                                  'id_option' => 'AR',
                                  'name' => 'Argentina'
                                ),
								array(
                                  'id_option' => 'CL',
                                  'name' => 'Chile'
                                ),
                                array(
                                  'id_option' => 'UY',
                                  'name' => 'Uruguay'
                                ),
                            ),
                            'id' => 'id_option',
                            'name' => 'name'
                        )
                    ),

                    array(
                        'col' => 3,
                        'type' => 'text',
                        'prefix' => '<i class="icon icon-envelope"></i>',
                        'desc' => $this->l('Enter a valid Contabilium.com username'),
                        'name' => 'CONTABILIUM_USERNAME',
                        'label' => $this->l('Username'),
                        'required' => true,
                    ),

                    array(
                        'col' => 3,
                        'type' => 'text',
                        'name' => 'CONTABILIUM_API_KEY',
                        'label' => $this->l('API Key'),
                        'desc' => $this->l('Enter a valid Contabilium.com API KEY'),
                        'required' => true,
                    ),

                    array(
                        'type' => 'html',
                        'name' => 'connection_status',
                        'label' => $this->l('Connection status'),
                        'html_content' =>
                            '<div style="line-height: 32px;">' .
                            ($this->contabilium->isConnected() ?
                                '<span class="label label-success">' . $this->l('Connected') . '</span>' . ((_PS_MODE_DEV_) ? '<br><pre>DEBUG<hr>Token: ' . $this->contabilium->getToken() . '</pre>' : '')
                                :
                                '<span class="label label-danger">' . $this->l('Not connected') . (($this->contabilium->hasError()) ? ' (' . $this->contabilium->getLastError() . ')' : '') . '</span>')
                            .'</div>'
                    ),

                    array(
                        'col' => 3,
                        'type' => 'text',
                        'name' => 'CONTABILIUM_INTEGRATION_ID',
                        'label' => $this->l('Integration ID'),
                        'desc' => $this->l('Enter a valid Contabilium.com Integration ID'),
                        'required' => true,
                    ),

                    array(
                        'type' => 'switch',
                        'label' => $this->l('Sync prices'),
                        'name' => 'CONTABILIUM_SYNC_PRICE',
                        'is_bool' => true,
                        'values' => array(
                            array(
                                'id' => 'active_on',
                                'value' => 1,
                                'label' => $this->l('Enabled')
                            ),
                            array(
                                'id' => 'active_off',
                                'value' => 0,
                                'label' => $this->l('Disabled')
                            )
                        ),
                        'js' => 'alert("hola");'
                    ),

                    array(
                        'type' => 'switch',
                        'label' => $this->l('Use price with tax'),
                        'name' => 'CONTABILIUM_USE_PRICE_WITH_TAX',
                        'is_bool' => true,
                        'disabled' => !Configuration::get('CONTABILIUM_SYNC_PRICE'),
                        'values' => array(
                            array(
                                'id' => 'active_on',
                                'value' => 1,
                                'label' => $this->l('Enabled')
                            ),
                            array(
                                'id' => 'active_off',
                                'value' => 0,
                                'label' => $this->l('Disabled')
                            )
                        ),
                    ),

                    array(
                        'type' => 'switch',
                        'label' => $this->l('Sync stock'),
                        'name' => 'CONTABILIUM_SYNC_STOCK',
                        'is_bool' => true,
                        'values' => array(
                            array(
                                'id' => 'active_on',
                                'value' => 1,
                                'label' => $this->l('Enabled')
                            ),
                            array(
                                'id' => 'active_off',
                                'value' => 0,
                                'label' => $this->l('Disabled')
                            )
                        ),
                    ),

                    array(
                        'type' => 'select',
                        'multiple' => true,
                        'name' => 'CONTABILIUM_ACCEPTED_STATUS[]',
                        'label' => $this->l('Accepted Order'),
                        'required' => true,
                        'options' => array(
                            'query' => $os,
                            'id' => 'id_option',
                            'name' => 'name'
                        )
                    ),
                    array(
                        'type' => 'select',
                        'multiple' => true,
                        'name' => 'CONTABILIUM_CANCELED_STATUS[]',
                        'label' => $this->l('Canceled Order'),
                        'required' => true,
                        'options' => array(
                            'query' => $os,
                            'id' => 'id_option',
                            'name' => 'name'
                        )
                    ),

                    array(
                        'type' => 'html',
                        'col' => 9,
                        'name' => 'CALLBACK_URL',
                        'label' => $this->l('Callback URL'),
                        'desc' => $this->l('URL to be set in Contabilium.com'),
                        'html_content' => '<div style="line-height: 24px; padding: 4px; border: 1px solid #efefef; font-family: courier;">' . $base_url . 'ring.php?token=' . Configuration::get('CONTABILIUM_CALLBACK_TOKEN') . '</div>'
                    ),

                    array(
                        'type' => 'html',
                        'col' => 9,
                        'name' => 'DOWNLOAD_LOG',
                        'label' => $this->l('Log File'),
                        'html_content' =>  '<a href="' . $base_url . 'cbl_log.php?token=' . Configuration::get('CONTABILIUM_CALLBACK_TOKEN') . '" target="_blank" class="btn btn-default btn-small">' . $this->l('Download now') . '</a>'
                    ),


                ),
                'submit' => array(
                    'title' => $this->l('Save'),
                ),
                'buttons' => array(),
            ),
        );

        if ($this->contabilium->isConnected()) {
            $form['form']['buttons'] = array(
                'sync-now' => array(
                    'title' => $this->l('Synchronize all products'),
                    'name' => 'submitContabiliumSyncNow',
                    'type' => 'submit',
                    'class' => 'btn btn-default btn-danger pull-left',
                    'icon' => 'process-icon-refresh',
                    'js' => "$(this).val('X');",
                ),
                'checkit-now' => array(
                    'title' => $this->l('Synchronize one product'),
                    'name' => 'submitContabiliumCheckItNow',
                    'type' => 'submit',
                    'class' => 'btn btn-default pull-left m-l-1',
                    'icon' => 'process-icon-cogs',
                    'js' => "code = prompt('" . $this->l('Please enter a valid reference code'). "', ''); if(code) { $(this).val('CS:'+code); } else { return false; }",
                ),
                'regenerate-token' => array(
                    'title' => $this->l('Regenerate Callback URL'),
                    'name' => 'submitContabiliumRegenerateToken',
                    'type' => 'submit',
                    'class' => 'btn btn-default pull-right',
                    'icon' => 'process-icon-reset',
                    'js' => "if(confirm('" . $this->l('Are you sure?') . ' '.  $this->l('You must notify CONTABILIUM about this change. Otherwise, integration will not work correctly.'). "')) { $(this).val('R'); } else { return false; }",
                ),
            );
        }

        return $form;
    }

    protected function getConfigFormValues()
    {
        return array(
            'CONTABILIUM_LIVE_MODE'          => Configuration::get('CONTABILIUM_LIVE_MODE'),
            'CONTABILIUM_USERNAME'           => Configuration::get('CONTABILIUM_USERNAME'),
            'CONTABILIUM_API_KEY'            => Configuration::get('CONTABILIUM_API_KEY'),
            'CONTABILIUM_INTEGRATION_ID'     => Configuration::get('CONTABILIUM_INTEGRATION_ID'),
            'CONTABILIUM_SYNC_PRICE'         => Configuration::get('CONTABILIUM_SYNC_PRICE'),
            'CONTABILIUM_USE_PRICE_WITH_TAX' => Configuration::get('CONTABILIUM_USE_PRICE_WITH_TAX'),
            'CONTABILIUM_SYNC_STOCK'         => Configuration::get('CONTABILIUM_SYNC_STOCK'),
            'CONTABILIUM_ACCEPTED_STATUS[]'  => unserialize(Configuration::get('CONTABILIUM_ACCEPTED_STATUS')),
            'CONTABILIUM_CANCELED_STATUS[]'  => unserialize(Configuration::get('CONTABILIUM_CANCELED_STATUS')),
            'CONTABILIUM_COUNTRY_CODE'       => Configuration::get('CONTABILIUM_COUNTRY_CODE'),
        );
    }

    protected function postProcessSave()
    {
        if (Shop::getContext() == Shop::CONTEXT_SHOP) {  // Solo grabo si estoy en el contexto de las tiendas.
            $form_values = $this->getConfigFormValues();

            foreach (array_keys($form_values) as $key) {
                if ($key == 'CALLBACK_URL') {
                    continue;
                }

                if (strpos($key, '[]') !== false) {
                    $key = str_replace('[]', '', $key);
                    Configuration::updateValue($key, serialize(Tools::getValue($key)));
                } else {
                    Configuration::updateValue($key, Tools::getValue($key));
                }
            }

            Configuration::updateValue('CONTABILIUM_TOKEN_CACHE', null);
        }
    }

    public function hookBackOfficeHeader()
    {
        if (Tools::getValue('module_name') == $this->name) {
            if (method_exists($this->context->controller, 'addJquery')) {
                $this->context->controller->addJquery();
                $this->context->controller->addJS($this->_path.'views/js/back.js');
            }
        }
    }

    public function hookActionOrderStatusPostUpdate($orderInfo)
    {
        $accepted = unserialize(Configuration::get('CONTABILIUM_ACCEPTED_STATUS'));
        $canceled = unserialize(Configuration::get('CONTABILIUM_CANCELED_STATUS'));

        $status_to_report = array_merge($accepted, $canceled);

        if (isset($orderInfo['newOrderStatus']) && in_array($orderInfo['newOrderStatus']->id, $status_to_report)) {
            $this->logToFile(
                sprintf(
                    "ORDER ID %s STATUS UPDATED --> NEW STATUS SET TO %s",
                    $orderInfo['id_order'],
                    $orderInfo['newOrderStatus']->id
                )
            );

            try {
                $ret = $this->contabilium->postOrder($orderInfo['id_order'], in_array($orderInfo['newOrderStatus']->id, $accepted))->getLastResult();

                $this->logToFile(
                    sprintf("ORDER ID %s REPORTED OK [response: %s]", $orderInfo['id_order'], var_export($ret, true))
                );
            } catch (Exception $e) {
                $this->logToFile(
                    sprintf("ORDER ID %s REPORTED WITH EXCEPTION: %s", $orderInfo['id_order'], $e->getMessage())
                );
            }
        }
    }

    public function hookActionOrderEdited($orderInfo)
    {
        $accepted = unserialize(Configuration::get('CONTABILIUM_ACCEPTED_STATUS'));
        $canceled = unserialize(Configuration::get('CONTABILIUM_CANCELED_STATUS'));

        $status_to_report = array_merge($accepted, $canceled);

        if (isset($orderInfo['order']) && in_array($orderInfo['order']->current_state, $status_to_report)) {
            $this->logToFile(sprintf("ORDER ID %s EDITED - MUST BE REPORTED", $orderInfo['order']->id));
            try {
                $ret = $this->contabilium->postOrder($orderInfo['order']->id, in_array($orderInfo['order']->current_state, $accepted))->getLastResult();

                $this->logToFile(
                    sprintf("ORDER ID %s REPORTED OK [response: %s]", $orderInfo['order']->id, var_export($ret, true))
                );
            } catch (Exception $e) {
                $this->logToFile(
                    sprintf("ORDER ID %s REPORTED WITH EXCEPTION: %s", $orderInfo['order']->id, $e->getMessage())
                );
            }
        }
    }

    public function logToFile($message)
    {
        if (_PS_MODE_DEV_) {
            file_put_contents(
                __DIR__ . '/contabilium.log',
                sprintf("[%s] %s\n", date('Y-m-d H:i:s'), $message),
                FILE_APPEND
            );
        }
    }

    public function hookDisplayAdminProductsExtra($params)
    {
        $this->context->smarty->assign(array(
            'id_product' => $params['id_product'],
            'rows' => $this->getRowsFromProductLog($params['id_product']),
        ));
        return $this->display(__FILE__, 'views/templates/hook/displayAdminProductsExtra.tpl');
    }

    public function hookDisplayAdminOrder($params)
    {
        $this->context->smarty->assign(array(
            'id_order' => $params['id_order'],
            'rows' => $this->getRowsFromOrderLog($params['id_order']),
        ));
        return $this->display(__FILE__, 'views/templates/hook/displayAdminOrder.tpl');
    }

    private function getRowsFromProductLog($id_product)
    {
        $sql =  ' SELECT l.*, coalesce(pl.name, \'--\') AS `product_name`, coalesce(pa.reference, \'--\') AS `attribute_reference` '
               .' FROM '._DB_PREFIX_.'contabilium_prod_update_log l '
               .'  LEFT JOIN '._DB_PREFIX_.'product_lang pl ON pl.id_product = l.id_product AND pl.id_lang = ' . $this->context->language->id
               .'  LEFT JOIN '._DB_PREFIX_.'product_attribute pa ON pa.id_product = l.id_product AND pa.id_product_attribute = l.id_product_attribute '
               .' WHERE l.id_product = ' . $id_product
               .' ORDER BY l.date_add DESC ';
        return Db::getInstance()->ExecuteS($sql);
    }

    private function getRowsFromOrderLog($id_order)
    {
        $sql =  ' SELECT *'
               .' FROM '._DB_PREFIX_.'contabilium_order_tx_log '
               .' WHERE id_order = ' . $id_order
               .' ORDER BY date_add DESC ';
        $data = array();
        foreach (Db::getInstance()->ExecuteS($sql) as $row) {
            $payload = json_decode(preg_replace('/[[:cntrl:]]/', '', $row['payload']));
            $data[] = array(
                'id_order' => $row['id_order'],
                'date_add' => $row['date_add'],
                'data'     => $row['payload'],
                'method'   => isset($payload->method) ? $payload->method : 'POST',
                'response' => isset($payload->response) ? $payload->response : 'N/D',
            );
        }
        return $data;
    }
}